@@ -1,13 +1,16 @@
+import json
 from collections import defaultdict
 from collections.abc import MutableMapping
 from dataclasses import dataclass
 from itertools import chain
-from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple
+from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple
 
 from dvc_objects.errors import ObjectFormatError
 from pygtrie import ShortKeyError  # noqa: F401, pylint: disable=unused-import
 from pygtrie import Trie
 
+from .hashfile.cache import Cache
+from .hashfile.hash_info import HashInfo
 from .hashfile.meta import Meta
 from .hashfile.tree import Tree, TreeError
 
@@ -15,7 +18,6 @@ if TYPE_CHECKING:
     from dvc_objects.fs.base import FileSystem
 
     from .hashfile.db import HashFileDB
-    from .hashfile.hash_info import HashInfo
     from .hashfile.obj import HashFile
 
 
@@ -36,6 +38,31 @@ class DataIndexEntry:
 
     loaded: Optional[bool] = None
 
+    @classmethod
+    def from_dict(cls, d: Dict[str, Dict]) -> "DataIndexEntry":
+        ret = cls()
+
+        meta = d.get("meta")
+        if meta:
+            ret.meta = Meta.from_dict(meta)
+
+        hash_info = d.get("hash_info")
+        if hash_info:
+            ret.hash_info = HashInfo.from_dict(hash_info)
+
+        return ret
+
+    def to_dict(self) -> Dict[str, Dict]:
+        ret = {}
+
+        if self.meta:
+            ret["meta"] = self.meta.to_dict()
+
+        if self.hash_info:
+            ret["hash_info"] = self.hash_info.to_dict()
+
+        return ret
+
 
 def _try_load(
     odbs: Iterable["HashFileDB"],
@@ -199,28 +226,40 @@ class DataIndex(MutableMapping):
         return self.traverse(node_factory, prefix=root_key)
 
 
-def _collect_dir(index, prefix, entry, path, fs):
+def _collect_dir(index, prefix, prefix_entry, path, fs, update=False):
+    dir_meta = Meta(nfiles=0, size=0, isdir=True)
+
     for root, dnames, fnames in fs.walk(path):
         sub_prefix = fs.path.relparts(root, path) if root != path else ()
         for name in chain(dnames, fnames):
             key = (*prefix, *sub_prefix, name)
             entry_path = fs.path.join(root, name)
+            entry = index.get(key)
+            if entry is None:
+                entry = DataIndexEntry()
+                index[key] = entry
+
+            entry.fs = fs
+            entry.path = entry_path
+            entry.cache = prefix_entry.cache
+            entry.remote = prefix_entry.remote
 
             # TODO: localfs.walk doesn't currently support detail=True,
             # so we have to call fs.info() manually
             meta = Meta.from_info(
                 fs.info(entry_path, refresh=True), fs.protocol
             )
-            index[key] = DataIndexEntry(
-                meta=meta,
-                fs=fs,
-                path=entry_path,
-                cache=entry.cache,
-                remote=entry.remote,
-            )
+            if entry.meta != meta and not update:
+                entry.hash_info = None
+
+            entry.meta = meta
+            dir_meta.nfiles += 1
+            dir_meta.size += meta.size
+
+    return dir_meta
 
 
-def collect(index, path, fs):
+def collect(index, path, fs, update=False):
     # NOTE: converting to list to avoid iterating and modifying the dict the
     # same time.
     items = list(index.iteritems(shallow=True))
@@ -230,22 +269,22 @@ def collect(index, path, fs):
         info = fs.info(entry_path, refresh=True)
 
         fs_meta = Meta.from_info(info, fs.protocol)
-        if entry.meta != fs_meta:
-            entry.meta = Meta.from_info(info, fs.protocol)
+        if entry.meta != fs_meta and not update:
             entry.hash_info = None
+        entry.meta = fs_meta
         entry.fs = fs
         entry.path = entry_path
 
         if info["type"] == "file":
             continue
 
-        _collect_dir(index, key, entry, entry_path, fs)
+        entry.meta = _collect_dir(index, key, entry, entry_path, fs)
 
 
 def md5(index):
-    from .hashfile.build import _build_file
+    from .hashfile.hash import file_md5
 
-    for _key, entry in index.iteritems():
+    for _, entry in index.iteritems():
         if entry.meta.isdir:
             continue
 
@@ -259,13 +298,7 @@ def md5(index):
         else:
             path = entry.path
 
-        _, obj = _build_file(
-            path,
-            entry.fs,
-            "md5",
-            odb=entry.cache,
-        )
-        entry.hash_info = obj.hash_info
+        entry.hash_info = HashInfo("md5", file_md5(path, entry.fs))
 
 
 def save(index):
@@ -379,3 +412,43 @@ def push(index):
 
 def fetch(index):
     transfer(index, "remote", "cache")
+
+
+def write_db(index, path):
+    cache = Cache(path)
+    with cache.transact():
+        for key, entry in index.iteritems():
+            cache["/".join(key)] = entry.to_dict()
+
+
+def read_db(path):
+    index = DataIndex()
+    cache = Cache(path)
+
+    with cache.transact():
+        for key in cache:
+            value = cache.get(key)
+            index[key.split("/")] = DataIndexEntry.from_dict(value)
+
+    return index
+
+
+def write_json(index, path):
+    with open(path, "w", encoding="utf-8") as fobj:
+        json.dump(
+            {
+                "/".join(key): entry.to_dict()
+                for key, entry in index.iteritems()
+            },
+            fobj,
+        )
+
+
+def read_json(path):
+    index = DataIndex()
+
+    with open(path, "r", encoding="utf-8") as fobj:
+        for key, value in json.load(fobj).items():
+            index[key.split("/")] = DataIndexEntry.from_dict(value)
+
+    return index
